{
   uLog, logging backend
   Copyright (C) 2008. Dejan Boras

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   Started On:    26.11.2008.
   Last Update:   14.02.2011.
}

{$IFDEF NO_LOGGING}
   {$FATAL uLog should not be used when NO_LOGGING is defined.}
{$ENDIF}

{$MODE OBJFPC}{$H+}{$I-}
UNIT uLog;

INTERFACE

   USES SysUtils, uStd
      {$IFDEF ANDROID}, androidlog{$ENDIF};

CONST
   {errors}
   logeNONE          = 00;
   logeGENERAL       = 01;
   logeNO_MEMORY     = 02;

   {file modes for log files | only writing modes are supported}
   logcREWRITE             = $00;
   logcAPPEND              = $01;

   {log type}
   logcNONE					= $0000;
   logcINFO					= $0001;
   logcWARNING				= $0002;
   logcERROR				= $0003;
   logcVERBOSE				= $0004;
   logcFATAL				= $0005;
   logcDEBUG				= $0006;

   {section}
   logcMAX_SECTIONS        = 20;
   logcMAX_SECTION_LEVEL   = logcMAX_SECTIONS-1;

   {logging settings}
   logcHandleLogs: boolean                = true;
   logcLogsInAppDir: boolean              = true;
   logcEmptyLineAfterHeader: boolean      = true;
   logcStartTimeDate: boolean             = true;
   logcEndTimeDate: boolean               = true;

   logcTag: string = 'uLog';

   {log file status indicators}
   logcINITIALIZED         = $01;
   logcOPENED              = $02;
   logcAPPEND_FAILED       = $04;
   logcERROR_STATE         = $08;
   logcOK                  = $10;

TYPE
   PLog = ^TLog;
   PLogHandler = ^TLogHandler;

   {a log handler}
   TLogHandler = record
      needopen: boolean;
      init: procedure(log: PLog);
      dispose: procedure(log: PLog);
      open: procedure(log: PLog);
      close: procedure(log: PLog);
      flush: procedure(log: PLog);
      writeln: procedure(log: PLog; const s: string);
      del: procedure(log: PLog);
   end;

   {a log file}
   TLog = record
      FileName, LogHeader: string;
      FMode, State: longword;
      {$IFNDEF ANDROID}f: ^text;{$ENDIF}

      SectionLevel: longint;

      {$IFDEF LOG_THREAD_SAFE}
      LogCS: TRTLCriticalSection;
      {$ENDIF}

      h: PLogHandler;
   end;

VAR
   logError: longint;
   {standard error logs}
   stdlog: TLog;

   logDummyHandler: TLogHandler;
   {$IFNDEF ANDROID}logStdHandler: TLogHandler;{$ENDIF}
   {$IFDEF ANDROID}logAndroidHandler: TLogHandler;{$ENDIF}
   logpDefaultHandler: PLogHandler = @logDummyHandler;

{ LOG FILE SUPPORT }
{initialize a TLog record}
procedure logInitRecord(var log: TLog);
{initialize a TLogHandler record}
procedure logInitRecord(var h: TLogHandler);

{creates a TLog on the heap and returns a pointer to it}
function logMake(): PLog;
{disposes a TLog record}
procedure logDispose(var log: TLog);
{disposes a PLog}
procedure logDispose(var log: PLog);

{initializes the log file, the second one reserves memory}
function logInitialize(var log: TLog; const filename, logheader: string; mode: longint): boolean;
{opens the log file}
procedure logOpen(var log: TLog);
{closes the log file}
procedure logClose(var log: TLog);
{deletes the log file}
procedure logDel(var log: TLog);
{de initializes the log file, the second one also frees memory}
procedure logDeInitialize(var log: TLog);
{verifies that the log is ok and ready for writting}
function logOk(var log: TLog): boolean;
{same as above, only it works with the standard logs}
function logOk(): boolean;
{Sets the log into a error state}
procedure logSetErrState(var log: TLog);

{Quickly initializes and deinitializes the standard log with the given filename}
procedure logInitStd(const filename, logheader: string; mode: longint);
procedure logDeInitStd();

{ LOG WRITTING }

{writes a string to the standard log file}
procedure logWriteln(var log: TLog; const s: string);
procedure logWriteln(const s: string);
procedure logWriteln(var log: TLog);
procedure logWriteln();
procedure logFlush(var log: TLog);
procedure logFlush();

{ SECTIONS }
procedure logEnterSection(var log: TLog; const s: string);
procedure logExitSection(var log: TLog);

procedure logEnterSection(const s: string);
procedure logExitSection();

IMPLEMENTATION

procedure logInitRecord(var log: TLog);
begin
   Zero(log, SizeOf(TLog));
   log.h := logpDefaultHandler;
end;

procedure logInitRecord(var h: TLogHandler);
begin
   Zero(h, SizeOf(h));
end;

function logMake(): PLog;
var
   log: PLog = nil;

begin
   new(log);
   if(log <> nil) then begin
      logInitRecord(log^); exit(log);
   end else begin
      logError := logeNO_MEMORY; exit(nil);
   end;
end;

procedure logDispose(var log: TLog);
begin
   log.FileName := '';

   if(log.h <> nil) then
      log.h^.dispose(@log);
end;

procedure logDispose(var log: PLog);
begin
   if(log <> nil) then begin
      logDispose(log^); dispose(log); log := nil;
   end;
end;

function logInitialize(var log: TLog; const filename, logheader: string; mode: longint): boolean;
begin
   result := false;

   {no filename provided, exit}
   if(logcHandleLogs) then
   if(filename <> '') and (log.state and logcINITIALIZED = 0) then begin
      {$IFDEF LOG_THREAD_SAFE}InitCriticalSection(log.LogCS);{$ENDIF}

      {store stuff}
      log.FileName := filename;
      log.LogHeader := logheader;
      log.FMode := mode;

      if(log.h <> nil) then
         log.h^.init(@log);

      if(logError = 0) then begin
         log.state := log.state or logcINITIALIZED;
         result := true;
      end;
   end;
end;

procedure logOpen(var log: TLog);
var
   td: TDateTime;

begin
   logError := logeNONE;
   if(log.h^.needopen) and (logcHandleLogs) then begin
      if(log.State and logcINITIALIZED > 0) then begin
         if(log.h <> nil) then
            log.h^.open(@log);
         if(log.State and logcERROR_STATE > 0) then exit;

         log.state := log.state or logcOPENED;
         log.state := log.state or logcOK;

         {write down the log header(if one exists)}
         if(log.LogHeader <> '') then begin
            logWriteln(log, log.LogHeader);
            {put an additional after the header(if chosen so)}
            if(logcEmptyLineAfterHeader = true) then
               logWriteln(log);
         end;

         if(logcStartTimeDate) then begin
            td := Now();
            logWriteln(log, 'Logging Start: '+DateToStr(td)+' at '+TimeToStr(td));
         end;
      end;
   end;
end;

procedure logClose(var log: TLog);
var
   td: TDateTime;

begin
   logError := logeNONE;

   if(logcHandleLogs) then
   if(log.state and logcOPENED > 0) then begin
      {write log end time}
      if(logcEndTimeDate) then begin
         td := Now();
         logWriteln('Logging End:   '+DateToStr(td)+' at '+TimeToStr(td));
      end;

      {close the file and set the state}
      if(log.h <> nil) then
         log.h^.close(@log);

      log.state := log.state xor logcOPENED;
      log.state := log.state and logcOK xor log.state;
   end;
end;

procedure logDel(var log: TLog);
begin
   logError := logeNONE;

   if(logcHandleLogs) then begin
      {close the log file}
      if(log.state and logcOPENED > 0) then begin
         logClose(log); if(IOResult <> 0) then exit;
       end;

       {erase the file}
       if(log.h <> nil) then
          log.h^.del(@log);
   end;
end;

procedure logDeInitialize(var log: TLog);
begin
   logError := logeNONE;

   if(logcHandleLogs <> false) then
   if(log.state and logcINITIALIZED > 0) then begin
      {$IFDEF LOG_THREAD_SAFE}DoneCriticalSection(log.LogCS);{$ENDIF}

      {if the file is opened try to close it}
      if(log.state and logcOPENED > 0) then begin
         logClose(log); if(logError <> logeNONE) then exit;
      end;

      log.state := log.state xor logcINITIALIZED;
      log.state := log.state and logcOK xor log.state;
   end;
end;

function logOk(var log: TLog): boolean;
begin
   {$IFDEF LOG_THREAD_SAFE}EnterCriticalSection(log.LogCS);{$ENDIF}
   logOk := (log.state and logcOK > 0);
   {$IFDEF LOG_THREAD_SAFE}LeaveCriticalSection(log.LogCS);{$ENDIF}
end;

function logOk(): boolean;
begin
   {$IFDEF LOG_THREAD_SAFE}EnterCriticalSection(stdlog.LogCS);{$ENDIF}
   logOk := (stdlog.state and logcOK > 0);
   {$IFDEF LOG_THREAD_SAFE}LeaveCriticalSection(stdlog.LogCS);{$ENDIF}
end;

procedure logSetErrState(var log: TLog);
begin
   {$IFDEF LOG_THREAD_SAFE}EnterCriticalSection(log.LogCS);{$ENDIF}
   logError := logeNONE;
   log.state := log.state xor logcERROR_STATE;
   log.state := log.state and logcOK xor log.state;
   {$IFDEF LOG_THREAD_SAFE}LeaveCriticalSection(log.LogCS);{$ENDIF}
end;

procedure logInitStd(const filename, logheader: string; mode: longint);
begin
   if(logcHandleLogs) then begin
      if(stdlog.state and logcINITIALIZED = 0) then begin
        {initialize the standard log file}
        logInitialize(stdlog, filename, logheader, mode);
        if(logError <> logeNONE) then exit;
        {open the standard log file}
        logOpen(stdlog);
      end;
   end;
end;

procedure logDeInitStd();
begin
   if(logcHandleLogs) then begin
      {close the standard log file}
      logClose(stdlog);
      if(logError <> logeNONE) then exit;
      {deinit the standard log file}
      logDeInitialize(stdlog);
      logDispose(stdlog);
   end;
end;

{ LOG WRITTING }
procedure logWriteln(var log: TLog; const s: string);
begin
   log.h^.writeln(@log, s);
end;

procedure logWriteln(const s: string);
begin
   logWriteln(stdlog, s);
end;

procedure logWriteln(var log: TLog);
begin
   log.h^.writeln(@log, '');
end;

procedure logWriteln();
begin
   logWriteln(stdlog);
end;

procedure logFlush(var log: TLog);
begin
   if(log.h <> nil) then
      log.h^.flush(@log);
end;

procedure logFlush();
begin
   logFlush(stdlog);
end;

{ SECTIONS }
procedure logEnterSection(var log: TLog; const s: string);
begin
   logWriteln(s);
   inc(log.SectionLevel);
   assert(log.SectionLevel < logcMAX_SECTIONS, 'Too many log sections, increase logcMAX_SECTIONS.');
end;

procedure logExitSection(var log: TLog);
begin
   dec(log.SectionLevel);
   assert(log.SectionLevel >= 0, 'Log section level must not be less than 0. An extra or premature logExitSection() call has been made.');
end;

procedure logEnterSection(const s: string);
begin
   logEnterSection(stdlog, s);
end;

procedure logExitSection();
begin
   logExitSection(stdlog);
end;

{ DUMMY LOG HANDLER}
procedure dummyopen(log: PLog);
begin
   if(log <> nil) then;
end;
procedure dummyproc(log: PLog);
begin
   if(log <> nil) then;
end;
procedure dummywriteln(log: PLog; const s: string);
begin
   if(log <> nil) then
      if(s <> '') then;
end;
procedure dummyflush(log: PLog);
begin
   if(log <> nil) then;
end;

{ STANDARD LOG HANDLER }
{$IFNDEF ANDROID}
procedure stdopen(log: PLog);
label
   repeatopen;

begin
   {jump here to try to open the file again, and who said labels are not good
   | other method could have been used but this one is simple}
repeatopen:

   {$IFNDEF ANDROID}
   {open the file in the correct mode.
   If the file cannot be appended try rewriting it}
   if(log^.fmode = logcAPPEND)then begin
      append(log^.f^);
      if(IOResult <> 0) then begin
         log^.state := log^.state or logcAPPEND_FAILED;
         log^.fmode := logcREWRITE;
         goto repeatopen;
      end;
   end else if(log^.fmode = logcREWRITE)then begin
      rewrite(log^.f^);
   end else begin
      log^.fmode := logcREWRITE; goto repeatopen;
   end;
   {$ENDIF}
end;

procedure stdinit(log: PLog);
begin
   {get memory for the file}
   new(log^.f);
   if(log^.f <> nil) then begin
      {set the inited bit(in state) to 1}
      log^.State := log^.State or logcINITIALIZED;

      {assign the filename to the file}
      Assign(log^.f^, log^.FileName);
   end else logError := logeNO_MEMORY;
end;

procedure stddispose(log: PLog);
begin
   if(log^.f <> nil) then
         dispose(log^.f);
   log^.f := nil;
end;

procedure stdclose(log: PLog);
begin
   close(log^.f^);
end;

VAR
   tabstr: string = #9#9#9#9#9+#9#9#9#9#9+#9#9#9#9#9+#9#9#9#9#9;

procedure stdwriteln(log: PLog; const s: string);
begin
   {$IFDEF LOG_THREAD_SAFE}
   EnterCriticalSection(log.LogCS);
   {$ENDIF}
   if(logOk(log^)) then begin
      if(log^.SectionLevel > 0) then
         writeln(log^.f^, copy(tabstr, 1, log^.SectionLevel)+s)
      else
         writeln(log^.f^, s);

      if(IOResult <> 0) then logSetErrState(log^);
   end;
   {$IFDEF LOG_THREAD_SAFE}
   LeaveCriticalSection(log^.LogCS);
   {$ENDIF}
end;

procedure stdflush(log: PLog);
begin
   {$IFDEF LOG_THREAD_SAFE}
   EnterCriticalSection(log^.LogCS);
   {$ENDIF}
   if(logOk(log^)) then begin
      flush(log^.f^);
      if(IOResult() <> 0) then logSetErrState(log^);
   end;
   {$IFDEF LOG_THREAD_SAFE}
   LeaveCriticalSection(log^.LogCS);
   {$ENDIF}
end;

procedure stddel(log: PLog);
begin
   erase(log^.f^);
   IOResult();
end;

procedure initStdH();
begin
   logStdHandler := logDummyHandler;
   logStdHandler.needopen := true;
   logStdHandler.open := @stdopen;
   logStdHandler.close := @stdclose;
   logStdHandler.init := @stdinit;
   logStdHandler.dispose := @stddispose;
   logStdHandler.writeln := @stdwriteln;
   logStdHandler.flush := @stdflush;

   logpDefaultHandler := @logStdHandler;
end;

{$ENDIF}

{ ANDROID LOG HANDLER }

{$IFDEF ANDROID}
CONST
	andrLogTypeRemaps: array[0..6] of longint = (
   {logcNONE}  		ANDROID_LOG_UNKNOWN
   {logcINFO}, 		ANDROID_LOG_INFO
   {logcWARNING}, 	ANDROID_LOG_WARN
   {logcERROR}, 	ANDROID_LOG_ERROR
   {logcVERBOSE}, 	ANDROID_LOG_VERBOSE
   {logcFATAL}, 	ANDROID_LOG_FATAL
   {logcDEBUG},		ANDROID_LOG_DEBUG);

procedure andrwriteln(log: PLog; const s: string);
var
	priority: longint;

begin
	priority := ANDROID_LOG_INFO;
	{if(priority >= 0) and (priority <= 6) then
		priority := andrLogTypeRemaps[p];}

   LOGI(ANDROID_LOG_INFO, PChar(logcTag), PChar(s));
end;

procedure initAndroidH();
begin
   logAndroidHandler := logDummyHandler;
   logAndroidHandler.writeln := @andrwriteln;

   logpDefaultHandler := @logAndroidHandler;
end;
{$ENDIF}

{ INITIALIZE }

procedure logInitHandlers();
begin
   logDummyHandler.needopen := false;
   logDummyHandler.open := @dummyopen;
   logDummyHandler.close := @dummyproc;
   logDummyHandler.init := @dummyproc;
   logDummyHandler.dispose := @dummyproc;
   logDummyHandler.writeln := @dummywriteln;
   logDummyHandler.flush := @dummyflush;
   logDummyHandler.del := @dummyproc;

   logpDefaultHandler := @logDummyHandler;

   {$IFNDEF ANDROID}initStdH();{$ENDIF}
   {$IFDEF ANDROID}initAndroidH();{$ENDIF}
end;

INITIALIZATION
   logInitHandlers();
   logInitRecord(stdlog);

FINALIZATION
   logDeInitStd();
END.

